from typing import Generator

import pytest

from solrstice import SelectQuery

from .helpers import (
    Config,
    create_config,
    index_test_data,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_get_response_gets_response(config: Config) -> None:
    name = "SelectGetResponse"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await index_test_data(config.context, name)

        builder = SelectQuery()
        solr_response = await builder.execute(config.context, name)
        docs_response = solr_response.get_docs_response()
        assert docs_response is not None
        assert docs_response.get_num_found() > 0
        assert docs_response.get_start() == 0
        assert len(docs_response.get_docs()) > 4
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_select_raw_gets_response(config: Config) -> None:
    name = "SelectRawGetResponse"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await index_test_data(config.context, name)

        builder = SelectQuery()
        solr_response = (await builder.execute_raw(config.context, name))["response"]
        assert solr_response["numFound"] > 0
        assert solr_response["start"] == 0
        docs_response = solr_response["docs"]
        assert len(docs_response) > 4
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_select_raw_blocking_gets_response(config: Config) -> None:
    name = "SelectRawBlockingGetResponse"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await index_test_data(config.context, name)

        builder = SelectQuery()
        solr_response = builder.execute_blocking_raw(config.context, name)["response"]
        assert solr_response["numFound"] > 0
        assert solr_response["start"] == 0
        docs_response = solr_response["docs"]
        assert len(docs_response) > 4
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_select_works_when_no_result(config: Config) -> None:
    name = "SelectNoResult"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await index_test_data(config.context, name)

        builder = SelectQuery(fq=["id:non_existent_id"])
        solr_response = await builder.execute(config.context, name)
        docs_response = solr_response.get_docs_response()
        assert docs_response is not None
        assert docs_response.get_num_found() == 0
        assert docs_response.get_start() == 0
        assert len(docs_response.get_docs()) == 0
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_select_works_with_cursor_mark(config: Config) -> None:
    name = "SelectCursorMark"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)
        cursor_mark = "*"
        current_iteration = 0
        while True:
            if current_iteration > 100:
                raise Exception("Cursor mark test failed. Too many iterations")
            builder = SelectQuery(
                fq=["age:[* TO *]"], rows=1, sort=["id desc"], cursor_mark=cursor_mark
            )
            result = await builder.execute(config.context, name)
            if result.get_next_cursor_mark() is not None:
                if cursor_mark == "*":
                    break
                cursor_mark = result.get_next_cursor_mark()  # type: ignore
            else:
                raise Exception("Cursor mark test failed. No next cursor mark")
            current_iteration += 1

    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_select_works_with_additional_params(config: Config) -> None:
    name = "SelectAdditionalParams"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await index_test_data(config.context, name)

        builder = SelectQuery(
            q="{!parent which=city_name:*}",
            fl=["id", "city_name", "child:[subquery]"],
            additional_params={
                "child.q": "*:*",
                "child.fl": ["id"],
                "child.rows": 1,
            },
        )
        solr_response = await builder.execute(config.context, name)
        docs_response = solr_response.get_docs_response()
        assert docs_response is not None
        assert docs_response.get_num_found() > 0
        child_result = docs_response.get_docs()[0].get("child")
        assert child_result is not None
        assert child_result["numFound"] > 0
        assert len(child_result["docs"]) == 1
    finally:
        await teardown_collection(config.context, name)



from typing import Generator

import pytest

from .helpers import Config, create_config, wait_for_solr, setup_collection, index_test_data, teardown_collection
from solrstice import SelectQuery, StatsComponent


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_stats_works(config: Config) -> None:
    name = "StatsWorks"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(stats=StatsComponent(fields=["age"]))
        response = await config.async_client.select(select_builder, name)
        stats = response.get_stats()
        assert stats is not None
        age_stats = stats.get_fields()["age"]
        assert age_stats.get_count() > 0
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_stats_works_string_field(config: Config) -> None:
    name = "StatsWorksStringField"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(stats=StatsComponent(fields=["id"]))
        response = await config.async_client.select(select_builder, name)
        stats = response.get_stats()
        assert stats is not None
        id_stats = stats.get_fields()["id"]
        assert id_stats.get_count() > 0
        assert id_stats.get_mean() is None
        assert isinstance(id_stats.get_min(), str)
        assert len(id_stats.get_min()) > 0
    finally:
        await teardown_collection(config.context, name)



from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from solrstice import SolrServerContext

async def create_collection(
    context: "SolrServerContext",
    name: str,
    config: str,
    shards: Optional[int] = 1,
    replication_factor: Optional[int] = 1,
) -> None:
    """
    Create a collection on the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to create.
    :param config: The name of the config to use for the collection.
    :param shards: The number of shards to create.
    :param replication_factor: The replication factor to use.
    """

def create_collection_blocking(
    context: "SolrServerContext",
    name: str,
    config: str,
    shards: Optional[int] = 1,
    replication_factor: Optional[int] = 1,
) -> None:
    """
    Create a collection on the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to create.
    :param config: The name of the config to use for the collection.
    :param shards: The number of shards to create.
    :param replication_factor: The replication factor to use.
    """

async def get_collections(context: "SolrServerContext") -> List[str]:
    """
    Get the list of collections on the Solr server.

    :param context: The Solr server context.
    :return: The list of collections on the Solr server.
    """

def get_collections_blocking(context: "SolrServerContext") -> List[str]:
    """
    Get the list of collections on the Solr server.

    :param context: The Solr server context.
    :return: The list of collections on the Solr server.
    """

async def collection_exists(context: "SolrServerContext", name: str) -> bool:
    """
    Check if a collection exists on the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to check.
    :return: True if the collection exists, False otherwise.
    """

def collection_exists_blocking(context: "SolrServerContext", name: str) -> bool:
    """
    Check if a collection exists on the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to check.
    :return: True if the collection exists, False otherwise.
    """

async def delete_collection(context: "SolrServerContext", name: str) -> None:
    """
    Delete a config from the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to delete.
    """

def delete_collection_blocking(context: "SolrServerContext", name: str) -> None:
    """
    Delete a config from the Solr server.

    :param context: The Solr server context.
    :param name: The name of the collection to delete.
    """

__all__ = [
    "create_collection",
    "create_collection_blocking",
    "get_collections",
    "get_collections_blocking",
    "collection_exists",
    "collection_exists_blocking",
    "delete_collection",
    "delete_collection_blocking",
]



import doctest
import os.path
import tempfile
from pathlib import Path
from sys import stderr
from typing import Dict, List

import mypy.api

import solrstice

from .helpers import find_mypy_config


def get_doctests_from_solrstice() -> Dict[str, List[doctest.Example]]:
    to_parse = [Path(solrstice.__path__[0])]  # type: ignore
    doctest_parser = doctest.DocTestParser()
    doctest_examples: Dict[str, List[doctest.Example]] = {}
    while to_parse:
        path = Path(to_parse.pop())
        for file in path.iterdir():
            if file.is_dir():
                to_parse.append(file)
            elif file.suffix in [".py", ".pyi"]:
                with open(file) as f:
                    doctests = doctest_parser.get_doctest(
                        f.read(), {}, str(file), str(file), 0
                    )
                    doctest_examples[file.name] = doctests.examples

    # Merge examples that are from the same code block
    for examples in doctest_examples.values():
        if examples:
            finished_examples = []
            current_example: doctest.Example = examples[0]
            current_lineno: int = current_example.lineno
            for example in examples[1:]:
                # Allow for a single whitespace between code blocks
                if example.lineno < current_lineno + 3:
                    current_example.source += example.source
                    current_lineno = example.lineno
                else:
                    finished_examples.append(current_example)
                    current_example = example
                    current_lineno = example.lineno
            examples.clear()
            examples.extend(finished_examples)
    return doctest_examples


def test_doctests_with_mypy() -> None:
    mypy_confg = find_mypy_config()
    solrstice_examples = get_doctests_from_solrstice()
    with tempfile.TemporaryDirectory() as d:
        for file, examples in solrstice_examples.items():
            file_path = os.path.join(d, file)
            with open(file_path, "w") as f:
                file_contents: List[str] = []
                # Add spacing so that the code is at the same line in the new file as it was in the original file
                for example in examples:
                    first_example_line = example.lineno
                    # Remove 1 to account for the extra line that will be added
                    while len(file_contents) < first_example_line - 1:
                        file_contents.append("\n")
                    file_contents.append(
                        f"def test_doctest_{example.lineno}() -> None:\n"
                    )
                    for line in example.source.split("\n"):
                        file_contents.append(f"    {line}\n")
                    file_contents.append("\n")
                f.write("".join(file_contents))
        results = list(mypy.api.run(["--config-file", f"{mypy_confg}", d]))
        # Replace file path with the original file name
        results[0] = results[0].replace(d, solrstice.__path__[0])  # type: ignore
        if results[2] != 0:
            print(f"\n{results[0]}", stderr)
            raise AssertionError("Mypy failed. See above for details.")



import re
import tempfile
from pathlib import Path
from sys import stderr
from typing import List, Tuple

import mypy.api

import solrstice

from .helpers import find_mypy_config


def find_readme_path() -> Path:
    solrstice_path = Path(solrstice.__path__[0])  # type: ignore
    solrstice_parent = solrstice_path.parent
    if (solrstice_parent / "README.md").exists():
        return solrstice_parent / "README.md"
    raise FileNotFoundError("Could not find README.md")


def extract_readme_examples() -> List[Tuple[int, str]]:
    """Extracts code examples from the README.md file with the first line of the example"""
    readme_path = find_readme_path()
    with open(readme_path) as f:
        readme_content = f.read()
    examples = []
    extraction_regex = r"```python\n(.*?)\n```"
    for match in re.finditer(extraction_regex, readme_content, re.DOTALL):
        start_position = match.start()
        line_number = readme_content.count("\n", 0, start_position) + 1
        examples.append((line_number, match.group(1)))
    return examples


def test_readme_with_mypy() -> None:
    mypy_config = find_mypy_config()
    readme_examples = extract_readme_examples()
    file_contents: List[str] = []
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py") as f:
        for example in readme_examples:
            while len(file_contents) < example[0] - 1:
                file_contents.append("\n")

            file_contents.append(f"def test_readme_example_{example[0]}() -> None:\n")
            for line in example[1].split("\n"):
                file_contents.append(f"    {line}\n")
        f.write("".join(file_contents))
        f.flush()
        results = list(mypy.api.run(["--config-file", f"{mypy_config}", f.name]))
        # Replace file path with the original file name
        results[0] = results[0].replace(f.name, str(find_readme_path()))  # type: ignore
        if results[2] != 0:
            print(f"\n{results[0]}", stderr)
            raise AssertionError("Mypy failed. See above for details.")



from typing import Any, Dict, List, Optional

__all__ = [
    "SolrFacetSetResult",
    "SolrStatsResult",
    "SolrStatsFieldResult",
    "SolrPivotFacetResult",
    "SolrFieldFacetResult",
    "SolrGroupFieldResult",
    "SolrGroupResult",
    "SolrJsonFacetResponse",
    "SolrDocsResponse",
    "SolrResponse",
]


class SolrFacetSetResult:
    """
    Gets the facet counts from a query
    """

    def get_queries(self) -> Dict[str, int]:
        """
        Gets the query facets

        :return: The query facets

        >>> from solrstice import AsyncSolrCloudClient, SolrServerContext, SelectQuery, FacetSetComponent, FacetSetComponent
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def query_select() -> None:
        ...     select_builder = SelectQuery(facet_set=FacetSetComponent(queries=["age:[0 TO 59]"]))
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_facet_set()
        ...     queries = facets.get_queries()
        ...     query = queries.get("age:[0 TO 59]")

        """
        pass

    def get_pivots(self) -> Dict[str, List["SolrPivotFacetResult"]]:
        """
        Gets the pivot facets

        :return: The pivot facets

        >>> from solrstice import FacetSetComponent, PivotFacetComponent, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def pivot_select() -> None:
        ...     select_builder = SelectQuery(facet_set=FacetSetComponent(pivots=PivotFacetComponent(["interests,age"])))
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_facet_set()
        ...     pivots = facets.get_pivots()
        ...     interests_age = pivots.get("interests,age")

        """
        pass

    def get_fields(self) -> Dict[str, List["SolrFieldFacetResult"]]:
        """
        Gets the field facets

        :return: The field facets

        >>> from solrstice import FacetSetComponent, FieldFacetComponent, FieldFacetEntry, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def field_facet_select() -> None:
        ...     facet_set = FacetSetComponent(fields=FieldFacetComponent(fields=[FieldFacetEntry("age")]))
        ...     select_builder = SelectQuery(facet_set=facet_set)
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_facet_set()
        ...     fields = facets.get_fields()
        ...     age = fields.get("age")

        """


class SolrPivotFacetResult:
    """
    Gets the pivot facet counts from a query
    """

    def get_value(self) -> Any:
        """
        Gets the value of the pivot

        :return: The value of the pivot
        """
        pass

    def get_pivots(self) -> List["SolrPivotFacetResult"]:
        """
        Gets additional pivots
        :return: The additional pivots
        """

    def get_queries(self) -> Dict[str, int]:
        """
        Gets the query facets
        :return: The query facets
        """

    def get_count(self) -> int:
        """
        Gets the count of the pivot
        :return: The count of the pivot
        """


class SolrFieldFacetResult:
    def get_key(self) -> Any:
        """
        Gets the key of the facet

        :return: The key of the facet
        """
        pass

    def get_count(self) -> int:
        """
        Gets the count of the facet
        :return: The count of the facet
        """


class SolrGroupFieldResult:
    """
    Represents a group field result
    """

    def get_group_value(self) -> Any:
        """
        Gets the group value
        :return: Group value
        """

    def get_doc_list(self) -> "SolrDocsResponse":
        """
        Gets the document response from solr
        :return: Document response
        """


class SolrGroupResult:
    """
    Represents a group result
    """

    def get_field_result(self) -> Optional[List[SolrGroupFieldResult]]:
        """
        Gets the field results form a group query
        :return: List of group field results

        >>> from solrstice import GroupingComponent, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def field_result() -> None:
        ...     group_builder = GroupingComponent(fields=["age"], limit=10)
        ...     select_builder = SelectQuery(fq=["age:[* TO *]"], grouping=group_builder)
        ...     groups = (await client.select(select_builder, "example_collection")).get_groups()
        ...     age_group = groups["age"]
        ...     docs = age_group.get_field_result()

        """

    def get_query_result(self) -> Optional["SolrDocsResponse"]:
        """
        Gets the query result from a group query
        :return: Query result

        >>> from solrstice import GroupingComponent, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def group_query_result() -> None:
        ...     group_builder = GroupingComponent(queries=["age:[0 TO 59]", "age:[60 TO *]"], limit=10)
        ...     select_builder = SelectQuery(fq=["age:[* TO *]"], grouping=group_builder)
        ...     groups = (await client.select(select_builder, "example_collection")).get_groups()
        ...     age_group = groups["age:[0 TO 59]"]
        ...     group = age_group.get_query_result()
        ...     assert group is not None
        ...     docs = group.get_docs()

        """

    def get_simple_result(self) -> Optional["SolrDocsResponse"]:
        """
        Gets the result from a group query where `GroupFormatting.Simple` was used
        :return: Simple result
        """

    def get_matches(self) -> int:
        """
        Gets the number of matches for a group query
        :return: Number of matches
        """

    def get_n_groups(self) -> int:
        """
        Gets the number of groups for a group query
        :return: Number of groups
        """


class SolrStatsFieldResult:
    def get_min(self) -> Any:
        """
        Gets the minimum value for this field
        :return: The minimum value
        """

    def get_max(self) -> Any:
        """
        Gets the maximum value for this field
        :return: The maximum value
        """

    def get_count(self) -> int:
        """
        Gets the count of values for this field
        :return: The count of values
        """

    def get_missing(self) -> int:
        """
        Gets the number of missing values for this field
        :return: The number of missing values
        """

    def get_sum(self) -> Optional[float]:
        """
        Gets the sum of values for this field
        :return: The sum of values
        """

    def get_mean(self) -> Optional[Any]:
        """
        Gets the mean of values for this field
        :return: The mean of values
        """

    def get_sum_of_squares(self) -> Optional[float]:
        """
        Gets the sum of squares for this field
        :return: The sum of squares
        """

    def get_stddev(self) -> Optional[float]:
        """
        Gets the standard deviation for this field
        :return: The standard deviation
        """


class SolrStatsResult:
    def get_fields(self) -> Dict[str, SolrStatsFieldResult]:
        """
        Gets the stats component fields from a query
        :return: The fields and their stats
        """


class SolrJsonFacetResponse:
    """
    A response from a json facet query
    """

    def get_val(self) -> Optional[Any]:
        """
        If a bucket facet, this value will be set
        :return: The value for this facet

        """

    def get_buckets(self) -> List["SolrJsonFacetResponse"]:
        """
        Get the buckets for this facet
        :return: A list of buckets

        >>> from solrstice import AsyncSolrCloudClient, SolrServerContext, SelectQuery, JsonFacetComponent, JsonTermsFacet
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def json_facet_buckets() -> None:
        ...     select_builder = SelectQuery(
        ...         json_facet=JsonFacetComponent(facets={"age": JsonTermsFacet("age")})
        ...     )
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_json_facets()
        ...     assert facets is not None
        ...     age_buckets = facets.get_nested_facets()["age"].get_buckets()
        ...     assert len(age_buckets) == 3

        """

    def get_num_buckets(self) -> Optional[int]:
        """
        Get the number of buckets for this facet, if specified in the query
        :return: The number of buckets
        """

    def get_missing(self) -> Optional[int]:
        """
        Get the number of missing values for this facet, if specified in the query
        :return: The number of missing values
        """

    def get_all_buckets(self) -> Optional[int]:
        pass

    def get_flat_facets(self) -> Dict[str, Any]:
        """
        Get stat counts for this facet
        :return: A dictionary of stat counts

        >>> from solrstice import JsonFacetComponent, JsonStatFacet, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        >>> async def json_flat_facets() -> None:
        ...     select_builder = SelectQuery(
        ...         json_facet=JsonFacetComponent(
        ...             facets={"total_people": JsonStatFacet("sum(count)")}
        ...         )
        ...     )
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_json_facets()
        ...     assert facets is not None
        ...     total_people = facets.get_flat_facets()["total_people"]
        ...     assert total_people == 1000

        """

    def get_nested_facets(self) -> Dict[str, "SolrJsonFacetResponse"]:
        """
        Get the nested facets for this facet
        :return: A dictionary of nested facets

        >>> from solrstice import JsonFacetComponent, JsonQueryFacet, SelectQuery, SolrServerContext, AsyncSolrCloudClient
        >>> async def main() -> None:
        ...     client = AsyncSolrCloudClient(SolrServerContext('localhost:8983'))
        ...     select_builder = SelectQuery(
        ...         json_facet=JsonFacetComponent(
        ...             facets={"below_60": JsonQueryFacet(q="age:[0 TO 59]")}
        ...         )
        ...     )
        ...     response = await client.select(select_builder, "example_collection")
        ...     facets = response.get_json_facets()
        ...     assert facets is not None
        ...     below_60 = facets.get_nested_facets().get("below_60")
        ...     assert below_60 is not None
        ...     assert below_60.get_count() == 4

        """

    def get_count(self) -> Optional[int]:
        """
        Get the count for this facet
        :return: The count for this facet
        """


class SolrDocsResponse:
    def get_num_found(self) -> int:
        """Get the number of documents found in the query"""

    def get_start(self) -> int:
        """Get the start index of the query"""

    def get_num_found_exact(self) -> bool:
        """Get whether the number of documents found is exact. This field only exists on Solr 8.6+. On older versions, this will always be true."""

    def get_docs(self) -> List[Dict[str, Any]]:
        """Get the documents from the query"""


class SolrResponse:
    """The response from a solr query"""

    def get_docs_response(self) -> Optional["SolrDocsResponse"]:
        """Get the response from a solr query"""

    def get_groups(self) -> Dict[str, "SolrGroupResult"]:
        """Get the groups from a solr query"""

    def get_next_cursor_mark(self) -> Optional[str]:
        """Get the next cursor mark from a solr query"""

    def get_facet_set(self) -> "SolrFacetSetResult":
        """Get facet counts"""

    def get_json_facets(self) -> Optional["SolrJsonFacetResponse"]:
        """Get json facets"""

    def get_stats(self) -> Optional["SolrStatsResult"]:
        """Get stats"""



import pickle

from solrstice import (
    CommitType,
    DeleteQuery,
    GroupingComponent,
    SelectQuery,
    UpdateQuery,
)


def test_pickle_works_select_query_builder() -> None:
    builder = SelectQuery(
        fq=["test", "test"],
        grouping=GroupingComponent(fields=["test"], main=True, facet=False),
    )
    string = pickle.dumps(builder)
    builder_copy: SelectQuery = pickle.loads(string)
    assert pickle.dumps(builder_copy) == string


def test_pickle_works_update_query_builder() -> None:
    builder = UpdateQuery(handler="test", commit_type=CommitType.Soft)
    string = pickle.dumps(builder)
    builder_copy: UpdateQuery = pickle.loads(string)
    assert pickle.dumps(builder_copy) == string


def test_pickle_works_delete_query_builder() -> None:
    builder = DeleteQuery()
    string = pickle.dumps(builder)
    builder_copy: DeleteQuery = pickle.loads(string)
    assert pickle.dumps(builder_copy) == string



import dataclasses
import random
from multiprocessing.pool import ThreadPool
from typing import Generator, Optional

import pytest

from solrstice import (
    BlockingSolrCloudClient,
    SolrBasicAuth,
    SolrServerContext,
    SolrSingleServerHost,
    UpdateQuery,
)

from .helpers import Config, create_config


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@dataclasses.dataclass
class PickableConfig:
    solr_host: str
    solr_username: Optional[str]
    solr_password: Optional[str]


def create_client(config: PickableConfig) -> BlockingSolrCloudClient:
    auth = (
        None
        if not config.solr_username
        else SolrBasicAuth(config.solr_username, config.solr_password)
    )
    return BlockingSolrCloudClient(
        SolrServerContext(SolrSingleServerHost(config.solr_host), auth)
    )


def index_independent(config: PickableConfig, collection_name: str) -> None:
    client = create_client(config)
    client.index(
        UpdateQuery(),
        collection_name,
        [{"id": str(random.randint(0, 10000000000))}],
    )


# TODO This does not work, and blocks forever
# def test_blocking_multiprocessing_works(config: Config):
#     name = "BlockingMultiprocessingWorks"
#
#     pickable_config = PickableConfig(config.solr_host, config.solr_username, config.solr_password)
#     client = create_client(pickable_config)
#     try:
#         client.delete_collection(name)
#     except:
#         pass
#     try:
#         client.delete_config(name)
#     except:
#         pass
#
#     client.upload_config(name, config.config_path)
#     client.create_collection(name, name, shards=1, replication_factor=1)
#
#     with Pool(processes=4) as pool:
#         tasks = [pool.apply_async(index_independent, (pickable_config, name)) for _ in range(10)]
#         [task.get(15) for task in tasks]
#
#     try:
#         client.delete_collection(name)
#     except:
#         pass
#     try:
#         client.delete_config(name)
#     except:
#         pass


def test_blocking_multithreading_works(config: Config) -> None:
    name = "BlockingMultithreadingWorks"

    pickable_config = PickableConfig(
        config.solr_host, config.solr_username, config.solr_password
    )
    client = create_client(pickable_config)
    try:
        client.delete_collection(name)
    except:
        pass
    try:
        client.delete_config(name)
    except:
        pass

    client.upload_config(name, config.config_path)
    client.create_collection(name, name, shards=1, replication_factor=1)

    with ThreadPool(processes=4) as pool:
        tasks = [
            pool.apply_async(index_independent, (pickable_config, name))
            for _ in range(10)
        ]
        [task.get(15) for task in tasks]

    try:
        client.delete_collection(name)
    except:
        pass
    try:
        client.delete_config(name)
    except:
        pass



from pathlib import Path

import solrstice


def find_mypy_config() -> Path:
    solrstice_path = Path(solrstice.__path__[0])  # type: ignore
    solrstice_parent = solrstice_path.parent
    if (solrstice_parent / "mypy.ini").exists():
        return solrstice_parent / "mypy.ini"
    raise FileNotFoundError("Could not find mypy.ini")



import asyncio
from typing import Generator

import pytest
from typing_extensions import Optional

from solrstice import (
    AsyncSolrCloudClient,
    BlockingSolrCloudClient,
    DeleteQuery,
    SelectQuery,
    SolrAuth,
    SolrBasicAuth,
    SolrServerContext,
    SolrSingleServerHost,
    UpdateQuery,
)

from .helpers import Config, create_config


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_async_client_works(config: Config) -> None:
    name = "AsyncClientWorks"

    client = AsyncSolrCloudClient(config.context)
    try:
        await client.delete_collection(name)
    except:
        pass
    try:
        await client.delete_config(name)
    except:
        pass

    await client.upload_config(name, config.config_path)
    await client.create_collection(name, name, shards=1, replication_factor=1)
    await client.index(UpdateQuery(), name, [{"id": "example_document"}])
    response = await client.select(SelectQuery(fq=["id:example_document"]), name)
    docs = response.get_docs_response()
    assert docs is not None
    assert docs.get_num_found() == 1

    response_raw = await client.select_raw(SelectQuery(fq=["id:example_document"]), name)
    assert response_raw["response"]["numFound"] == 1

    await client.delete(DeleteQuery(ids=["example_document"]), name)
    response = await client.select(SelectQuery(fq=["id:example_document"]), name)
    docs = response.get_docs_response()
    assert docs is not None
    assert docs.get_num_found() == 0

    await client.delete_collection(name)
    await client.delete_config(name)


def test_blocking_client_works(config: Config) -> None:
    name = "BlockingClientWorks"

    client = BlockingSolrCloudClient(config.context)
    try:
        client.delete_collection(name)
    except:
        pass
    try:
        client.delete_config(name)
    except:
        pass

    client.upload_config(name, config.config_path)
    client.create_collection(name, name, shards=1, replication_factor=1)
    client.index(UpdateQuery(), name, [{"id": "example_document"}])
    response = client.select(SelectQuery(fq=["id:example_document"]), name)
    docs = response.get_docs_response()
    assert docs is not None
    assert docs.get_num_found() == 1

    response_raw = client.select_raw(SelectQuery(fq=["id:example_document"]), name)
    assert response_raw["response"]["numFound"] == 1

    client.delete(DeleteQuery(ids=["example_document"]), name)
    response = client.select(SelectQuery(fq=["id:example_document"]), name)
    docs = response.get_docs_response()
    assert docs is not None
    assert docs.get_num_found() == 0

    client.delete_collection(name)
    client.delete_config(name)


@pytest.mark.asyncio
async def test_multiple_clients_works() -> None:
    name = "MultipleClientWorks"

    config_1 = create_config()
    config_2 = create_config()

    client = AsyncSolrCloudClient(config_1.context)
    client_2 = AsyncSolrCloudClient(config_2.context)

    try:
        await client.delete_config(name)
    except:
        pass

    await client.upload_config(name, config_1.config_path)

    results = await asyncio.gather(*[client.get_configs(), client_2.get_configs()])
    assert name in results[0]
    assert name in results[1]

    await client.delete_config(name)


@pytest.mark.asyncio
async def test_subclassing_client_works() -> None:
    class SolrClient(AsyncSolrCloudClient):
        def __new__(cls, host: str, auth: Optional[SolrAuth] = None):  # type: ignore
            context = SolrServerContext(SolrSingleServerHost(host), auth)
            return super().__new__(cls, context=context)  # type: ignore

        def test_method(self) -> str:
            return "test"

    name = "SubclassingClientWorks"

    config = create_config()

    client = SolrClient(
        config.solr_host,
        (
            SolrBasicAuth(config.solr_username, config.solr_password)
            if (config.solr_username and config.solr_password)  # type: ignore
            else None
        ),
    )  # type: ignore

    try:
        await client.delete_config(name)
    except:
        pass

    await client.upload_config(name, config.config_path)
    assert client.test_method() == "test"

    await client.delete_config(name)



from typing import Generator

import pytest

from solrstice.config import (
    config_exists,
    config_exists_blocking,
    delete_config,
    delete_config_blocking,
    upload_config,
    upload_config_blocking,
)

from .helpers import Config, create_config


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_config_all_async_functions_exported(config: Config) -> None:
    try:
        await delete_config(config.context, "UploadConfig")
    except RuntimeError:
        pass
    assert not await config_exists(config.context, "UploadConfig")
    await upload_config(
        config.context,
        "UploadConfig",
        config.config_path,
    )
    assert await config_exists(config.context, "UploadConfig")
    await delete_config(config.context, "UploadConfig")


def test_config_all_blocking_functions_exported(config: Config) -> None:
    try:
        delete_config_blocking(config.context, "UploadConfig")
    except RuntimeError:
        pass
    assert not config_exists_blocking(config.context, "UploadConfig")
    upload_config_blocking(
        config.context,
        "UploadConfig",
        config.config_path,
    )
    assert config_exists_blocking(config.context, "UploadConfig")
    delete_config_blocking(config.context, "UploadConfig")



from typing import Generator

import pytest

from solrstice import DismaxQuery, EdismaxQuery, LuceneQuery, SelectQuery

from .helpers import (
    Config,
    create_config,
    index_test_data,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_lucene_query_parser(config: Config) -> None:
    name = "LuceneQueryParser"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        query_parser = LuceneQuery(df="population")
        select_builder = SelectQuery(q="outdoors", def_type=query_parser)
        (await select_builder.execute(config.context, name)).get_docs_response()
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_dismax_query_parser(config: Config) -> None:
    name = "DismaxQueryParser"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        query_parser = DismaxQuery(qf="interests^20", bq=["interests:cars^20"])
        select_builder = SelectQuery(q="outdoors", def_type=query_parser)
        response = (
            await select_builder.execute(config.context, name)
        ).get_docs_response()
        assert response is not None
        first_doc = response.get_docs()[0]
        assert first_doc["id"] == "city_Alta_20"
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_edismax_query_parser(config: Config) -> None:
    name = "EdismaxQueryParser"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        query_parser = EdismaxQuery(qf="interests^20", bq=["interests:cars^20"])
        select_builder = SelectQuery(q="outdoors", def_type=query_parser)
        response = (
            await select_builder.execute(config.context, name)
        ).get_docs_response()
        assert response is not None
        first_doc = response.get_docs()[0]
        assert first_doc["id"] == "city_Alta_20"
    finally:
        await teardown_collection(config.context, name)



from typing import Generator

import pytest

from solrstice import CommitType, UpdateQuery

from .helpers import (
    Config,
    create_config,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_index_indexes_documents(config: Config) -> None:
    name = "IndexIndexesDocuments"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)

        await UpdateQuery(handler="update", commit_type=CommitType.Soft).execute(
            config.context, name, [{"id": "test"}]
        )
    finally:
        await teardown_collection(config.context, name)



from typing import Generator

import pytest

from solrstice import SelectQuery

from .helpers import (
    Config,
    create_config,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_syntax_error_returns_sensible_error(config: Config) -> None:
    name = "SyntaxError"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        with pytest.raises(Exception) as e:
            await config.async_client.select(
                SelectQuery(fq=["some_field::syntax_error"]), name
            )
            assert "400" in str(e)
    finally:
        await teardown_collection(config.context, name)



import asyncio
import logging
from typing import Generator

import pytest
from _pytest.logging import LogCaptureFixture

from solrstice import AsyncSolrCloudClient, OffLoggingPolicy, SolrServerContext

from .helpers import Config, create_config, wait_for_solr


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config(logging=True)


log_lock = asyncio.Lock()


# TODO This test fails if run in parallel with the rest of the test suite, but not if run alone
# @pytest.mark.asyncio
# async def test_logging_logs_message(config: Config, caplog: LogCaptureFixture) -> None:
#     wait_for_solr(config.solr_host, 30)
#
#     async with log_lock:
#         with caplog.at_level(logging.DEBUG):
#             caplog.clear()
#             assert not any(
#                 "Sending Solr request to" in msg
#                 for msg in [x.getMessage() for x in caplog.records]
#             ), "Logs are not empty"
#             await config.async_client.get_configs()
#             assert any(
#                 "Sending Solr request to" in msg
#                 for msg in [x.getMessage() for x in caplog.records]
#             ), "Expected log message not found"


@pytest.mark.asyncio
async def test_logging_does_not_log_message_if_disabled(
    config: Config, caplog: LogCaptureFixture
) -> None:
    wait_for_solr(config.solr_host, 30)

    context = SolrServerContext(config.solr_host, config.solr_auth, OffLoggingPolicy())
    client = AsyncSolrCloudClient(context)

    async with log_lock:
        with caplog.at_level(logging.DEBUG):
            caplog.clear()
            assert not any(
                "Sending Solr request to" in msg
                for msg in [x.getMessage() for x in caplog.records]
            ), "Logs are not empty"
            await client.get_configs()
            assert not any(
                "Sending Solr request to" in msg
                for msg in [x.getMessage() for x in caplog.records]
            ), "Logs are not empty"


# @pytest.mark.asyncio
# async def test_logging_works_if_request_fails(
#     config: Config, caplog: LogCaptureFixture
# ) -> None:
#     name = "LoggingWorksIfRequestFails"
#     wait_for_solr(config.solr_host, 30)
#     async with log_lock:
#         with caplog.at_level(logging.DEBUG):
#             try:
#                 await setup_collection(config.context, name, config.config_path)
#                 await index_test_data(config.context, name)
#
#                 caplog.clear()
#                 assert not any(
#                     "Sending Solr request to" in msg
#                     for msg in [x.getMessage() for x in caplog.records]
#                 ), "Logs are not empty"
#                 try:
#                     await config.async_client.select(
#                         SelectQuery(fq=["this_is_a_syntax_error::0"]), name
#                     )
#                 except Exception:
#                     pass
#                 assert any(
#                     "Sending Solr request to" in msg
#                     for msg in [x.getMessage() for x in caplog.records]
#                 ), "Expected log message not found"
#             finally:
#                 await teardown_collection(config.context, name)



from typing import Generator

import pytest

from solrstice.collection import (
    collection_exists,
    collection_exists_blocking,
    create_collection,
    create_collection_blocking,
    delete_collection,
    delete_collection_blocking,
)
from solrstice.config import (
    config_exists,
    config_exists_blocking,
    delete_config,
    delete_config_blocking,
    upload_config,
    upload_config_blocking,
)

from .helpers import Config, create_config


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_collection_all_async_functions_exported(config: Config) -> None:
    name = "CollectionConfig"

    try:
        await delete_collection(config.context, name)
    except RuntimeError:
        pass
    try:
        await delete_config(config.context, name)
    except RuntimeError:
        pass
    assert not await collection_exists(config.context, name)
    assert not await config_exists(config.context, name)
    await upload_config(
        config.context,
        name,
        config.config_path,
    )
    await create_collection(config.context, name, name, 1, 1)
    assert await collection_exists(config.context, name)
    await delete_collection(config.context, name)
    await delete_config(config.context, name)


def test_collection_all_blocking_functions_exported(config: Config) -> None:
    name = "CollectionBlockingConfig"

    try:
        delete_collection_blocking(config.context, name)
    except RuntimeError:
        pass
    try:
        delete_config_blocking(config.context, name)
    except RuntimeError:
        pass
    assert not collection_exists_blocking(config.context, name)
    assert not config_exists_blocking(config.context, name)
    upload_config_blocking(
        config.context,
        name,
        config.config_path,
    )
    create_collection_blocking(config.context, name, name, 1, 1)
    assert collection_exists_blocking(config.context, name)
    delete_collection_blocking(config.context, name)
    delete_config_blocking(config.context, name)



import abc
from abc import ABC
from enum import Enum
from os import PathLike
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

__all__ = [
    "SolrAuth",
    "SolrBasicAuth",
    "QueryOperator",
    "DefType",
    "LuceneQuery",
    "DismaxQuery",
    "EdismaxQuery",
    "FacetSetComponent",
    "StatsComponent",
    "PivotFacetComponent",
    "FieldFacetComponent",
    "FieldFacetSort",
    "FieldFacetMethod",
    "FieldFacetEntry",
    "JsonFacetComponent",
    "JsonFacetType",
    "JsonFacetSortDirection",
    "JsonTermsFacet",
    "JsonTermsFacetMethod",
    "JsonQueryFacet",
    "JsonStatFacet",
    "SolrHost",
    "SolrSingleServerHost",
    "SolrMultipleServerHost",
    "ZookeeperEnsembleHost",
    "ZookeeperEnsembleHostConnector",
    "LoggingPolicy",
    "OffLoggingPolicy",
    "FastLoggingPolicy",
    "PrettyLoggingPolicy",
    "SolrServerContext",
    "GroupFormatting",
    "GroupingComponent",
    "SelectQuery",
    "CommitType",
    "UpdateQuery",
    "DeleteQuery",
    "AsyncSolrCloudClient",
    "BlockingSolrCloudClient",
]

if TYPE_CHECKING:
    from solrstice.models import SolrResponse


# region auth
class SolrAuth(ABC):
    """
    Base class for Solr authentication
    Valid implementations are :class:`SolrBasicAuth`
    """


class SolrBasicAuth(SolrAuth):
    """Basic authentication for Solr

    :param username: Username for Solr
    :param password: Password for Solr
    """

    def __init__(self, username: str, password: Optional[str] = None) -> None:
        pass


# endregion

# region def_type
class QueryOperator(Enum):
    """
    The default query operator
    """

    AND = "AND"
    OR = "OR"


class DefType(abc.ABC):
    """
    Specify query type.
    Valid implementations are :class:`LuceneQuery`, :class:`DismaxQuery`, and :class:`EdismaxQuery`
    """

    pass


class LuceneQuery(DefType):
    """
    Create a Lucene query builder.

    :param q_op: Query operator.
    :param df: Default field
    :param sow: Split on whitespace
    """

    def __init__(
            self,
            q_op: Optional[QueryOperator] = None,
            df: Optional[str] = None,
            sow: Optional[bool] = None,
    ):
        pass


class DismaxQuery(DefType):
    """
    Create a DisMax query builder.

    :param q_alt: Alternate query
    :param qf: Query fields
    :param mm: Minimum match
    :param pf: Phrase fields
    :param ps: Phrase slop
    :param qs: Query slop
    :param tie: Tie breaker
    :param bq: Boost query
    :param bf: Boost functions
    """

    def __init__(
            self,
            q_alt: Optional[str] = None,
            qf: Optional[str] = None,
            mm: Optional[str] = None,
            pf: Optional[str] = None,
            ps: Optional[str] = None,
            qs: Optional[str] = None,
            tie: Optional[str] = None,
            bq: Optional[List[str]] = None,
            bf: Optional[List[str]] = None,
    ):
        pass


class EdismaxQuery(DefType):
    """
    Create an Edismax query builder.

    :param q_alt: Alternate query
    :param qf: Query fields
    :param mm: Minimum match
    :param mm_auto_relax: Automatically relax mm
    :param pf: Phrase fields
    :param pf2: Phrase fields 2
    :param pf3: Phrase fields 3
    :param ps: Phrase slop
    :param ps2: Phrase slop 2
    :param ps3: Phrase slop 3
    :param qs: Query slop
    :param tie: Tie breaker
    :param bq: Boost query
    :param bf: Boost functions
    :param sow: Split on whitespace
    :param boost: Boost
    :param lowercase_operators: Lowercase operators
    :param stopwords: Stopwords
    :param uf: User fields
    """

    def __init__(
            self,
            q_alt: Optional[str] = None,
            qf: Optional[str] = None,
            mm: Optional[str] = None,
            mm_auto_relax: Optional[bool] = None,
            pf: Optional[str] = None,
            pf2: Optional[str] = None,
            pf3: Optional[str] = None,
            ps: Optional[str] = None,
            ps2: Optional[str] = None,
            ps3: Optional[str] = None,
            qs: Optional[str] = None,
            tie: Optional[str] = None,
            bq: Optional[List[str]] = None,
            bf: Optional[List[str]] = None,
            sow: Optional[bool] = None,
            boost: Optional[List[str]] = None,
            lowercase_operators: Optional[bool] = None,
            stopwords: Optional[bool] = None,
            uf: Optional[str] = None,
    ):
        pass


# endregion

# region facet_set

class FacetSetComponent:
    """
    Creates a facet set component allowing for counting of facets of different types

    :param queries: A list of queries to facet on
    :param fields: A list of fields to facet on
    :param pivots: A list of pivots to facet on
    """

    def __init__(
            self,
            queries: Optional[List[str]] = None,
            fields: Optional["FieldFacetComponent"] = None,
            pivots: Optional["PivotFacetComponent"] = None,
    ):
        pass


class PivotFacetComponent:
    """
    Allows faceting using pivots

    :param pivots: A list of pivots to facet on
    :param min_count: The minimum count for a facet to be returned
    """

    def __init__(self, pivots: List[str], min_count: Optional[str] = None):
        pass


class FieldFacetComponent:
    """
    Allows faceting using fields

    :param fields: A list of fields to facet on
    :param exclude_terms: Comma separated list of terms to exclude from the facet. Escaping needs to be done manually
    """

    def __init__(
            self,
            fields: List["FieldFacetEntry"],
            exclude_terms: Optional[str] = None,
    ):
        pass


class FieldFacetSort(Enum):
    """
    The sort order for a field facet
    """

    Count = "Count"
    Index = "Index"


class FieldFacetMethod(Enum):
    """
    The method for a field facet
    """

    Enum = "Enum"
    Fc = "Fc"
    Fcs = "Fcs"


class FieldFacetEntry:
    """

    :param field: The field to facet on
    :param prefix: Limit the terms to those starting with the given prefix
    :param contains: Limit the terms to those containing the given substring
    :param contains_ignore_case: Whether to ignore case when filtering by contains
    :param sort: The sort order for the facet
    :param limit: The maximum number of facet entries to return
    :param offset: The offset into the facet entries to return
    :param min_count: The minimum count needed for a facet to be returned
    :param missing: Whether to include a facet for missing values
    :param method: The method to use for the facet. Default is Fc
    :param enum_cache_min_df: The minimum document frequency for a term to be included in the facet cache. Only used for Enum method
    :param exists: Cap facet counts by 1
    """

    def __init__(
            self,
            field: str,
            prefix: Optional[str] = None,
            contains: Optional[str] = None,
            contains_ignore_case: Optional[bool] = None,
            sort: Optional["FieldFacetSort"] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            min_count: Optional[int] = None,
            missing: Optional[bool] = None,
            method: Optional["FieldFacetMethod"] = None,
            enum_cache_min_df: Optional[int] = None,
            exists: Optional[bool] = None,
    ):
        pass


# endregion

# region stats
class StatsComponent:
    """
    Stats component, gets statistics on fields

    :param fields: A list of fields to get statistics on
    """

    def __init__(self, fields: Optional[List[str]] = None):
        pass


# region json_facet

class JsonFacetComponent:
    """
    Json faceting component

    :param facets: A dictionary of facets to apply to the query
    """

    def __init__(self, facets: Optional[Dict[str, "JsonFacetType"]] = None):
        pass


class JsonFacetSortDirection(Enum):
    """
    The sort direction for a json facet
    """
    Asc = "Asc"
    Desc = "Desc"


class JsonFacetType(abc.ABC):
    """
    Base class for a json facet type
    Valid implementations are :class:`JsonTermsFacet`, :class:`JsonQueryFacet`, and :class:`JsonStatFacet`
    """

    pass


class JsonTermsFacetMethod(Enum):
    DocValues = "DocValues"
    UnInvertedField = "UnInvertedField"
    DocValuesHash = "DocValuesHash"
    Enum = "Enum"
    Stream = "Stream"
    Smart = "Smart"


class JsonTermsFacet(JsonFacetType):
    """
    Do a terms facet on a field

    :param field: The field to facet on
    :param offset: The offset into the list of terms
    :param limit: The maximum number of terms to return
    :param sort: The sort order for the terms
    :param prelim_sort: More performant approximation of the sort order for the terms
    :param overrequest:
    :param refine:
    :param overrefine:
    :param mincount: The minimum count for a term to be returned
    :param missing: Return a bucket for missing values
    :param num_buckets: Return the number of buckets
    :param all_buckets:
    :param prefix: Limit the terms to those starting with the given prefix
    :param facets: A dictionary of sub-facets to apply to the query
    :param method: The method to use for the facet.
    """

    def __init__(
            self,
            field: str,
            offset: Optional[int] = None,
            limit: Optional[int] = None,
            sort: Optional[Dict[str, JsonFacetSortDirection]] = None,
            prelim_sort: Optional[Dict[str, JsonFacetSortDirection]] = None,
            overrequest: Optional[int] = None,
            refine: Optional[bool] = None,
            overrefine: Optional[bool] = None,
            mincount: Optional[int] = None,
            missing: Optional[bool] = None,
            num_buckets: Optional[bool] = None,
            all_buckets: Optional[bool] = None,
            prefix: Optional[str] = None,
            facets: Optional[Dict[str, JsonFacetType]] = None,
            method: Optional[JsonTermsFacetMethod] = None,
    ):
        pass


class JsonQueryFacet(JsonFacetType):
    """
    Do a query facet

    :param q: The query to facet on
    :param limit: The maximum number of terms to return
    :param offset: The offset into the list of terms
    :param sort: The sort order for the terms
    :param fq: A list of filters to apply to the query
    :param facets: A list of sub-facets to apply to the query
    """

    def __init__(
            self,
            q: Optional[str] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            sort: Optional[str] = None,
            fq: Optional[str] = None,
            facets: Optional[Dict[str, JsonFacetType]] = None,
    ):
        pass


class JsonStatFacet(JsonFacetType):
    """
    Do a stat facet

    :param query: The query to facet on
    """

    def __init__(self, query: str):
        pass


# endregion

# region hosts
class SolrHost(ABC):
    """
    Base class for Solr hosts
    Valid implementations are :class:`SolrSingleServerHost`, :class:`SolrMultipleServerHost`, and :class:`ZookeeperEnsembleHost`
    """


class SolrSingleServerHost(SolrHost):
    """Solr host for a single Solr instance

    :param host: Hostname of the Solr instance
    """

    def __init__(self, host: str) -> None:
        pass


class SolrMultipleServerHost(SolrHost):
    """Solr host for multiple solr instances

    :param hosts: List of Solr instances
    :param timeout: Amount of seconds before declaring a node not responding, and going to the next
    """

    def __init__(self, hosts: List[str], timeout: float) -> None:
        pass


class ZookeeperEnsembleHost(SolrHost):
    """Zookeeper ensemble connection. Cannot be constructed directly, use ZookeeperEnsembleHostConnector instead"""


class ZookeeperEnsembleHostConnector:
    """The builder for a Zookeeper ensemble host

    :param hosts: List of Zookeeper instances
    :param timeout: Timeout for connecting to Zookeeper
    """

    def __init__(self, hosts: List[str], timeout: float) -> None:
        pass

    async def connect(self) -> ZookeeperEnsembleHost:
        """Connect to the Zookeeper ensemble"""
        pass

    def connect_blocking(self) -> ZookeeperEnsembleHost:
        """Connect to the Zookeeper ensemble"""
        pass


class LoggingPolicy(abc.ABC):
    """
    Policy describing how to log solr queries.
    Valid values are :class:`OffLoggingPolicy`, :class:`FastLoggingPolicy`, and :class:`PrettyLoggingPolicy`
    """

    pass


class OffLoggingPolicy(LoggingPolicy):
    """Do not log requests"""

    def __init__(self) -> None:
        pass


class FastLoggingPolicy(LoggingPolicy):
    """For each request create a logging::DEBUG message with url, headers, and body

    :param max_body_length: How long to allow the body to be before dropping to log it
    """

    def __init__(self, max_body_length: int) -> None:
        pass


class PrettyLoggingPolicy(LoggingPolicy):
    """For each request create a logging::DEBUG message with url, headers, and a pretty body

    :param max_body_length: How long to allow the body to be before dropping to log it
    """

    def __init__(self, max_body_length: int) -> None:
        pass


class SolrServerContext:
    """The context for a connection to a solr instance

    :param host: An instance of SolrHost specifying how to connect to a solr instance. If given as a string it creates a :class:`SolrSingleServerHost`
    :param auth: An instance of SolrAuth specifying how to authenticate with the solr instance
    :param logging_policy: How to log solr queries, valid values are :class:`OffLoggingPolicy`, :class:`FastLoggingPolicy`, and :class:`PrettyLoggingPolicy`
    """

    def __init__(
            self,
            host: Union[SolrHost, str],
            auth: Optional[SolrAuth] = None,
            logging_policy: Optional[LoggingPolicy] = None,
    ):
        pass


# endregion

# region group
class GroupFormatting(Enum):
    Simple = "Simple"
    Grouped = "Grouped"


class GroupingComponent:
    """
    Grouping component, used in conjunction with SelectQuery

    :param fields: Fields to group results by
    :param queries: Queries to group by
    :param limit: Limit the number of groups returned for each set of grouped documents
    :param offset: Offset the number of groups returned for each set of grouped documents
    :param sort: Sort the groups
    :param format: The group format, either Simple, or Grouped
    :param main: Should the group result be the main result
    :param n_groups: Should the number of groups be counted
    :param truncate: Truncate
    :param facet: Facet
    """

    def __init__(
            self,
            fields: Optional[List[str]] = None,
            queries: Optional[List[str]] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            sort: Optional[List[str]] = None,
            format: Optional[GroupFormatting] = None,
            main: Optional[bool] = None,
            n_groups: Optional[bool] = None,
            truncate: Optional[bool] = None,
            facet: Optional[bool] = None,
    ):
        pass


# endregion

# region queries
class SelectQuery:
    """Builder for a select query

    :param q: The query string
    :param fq: The filter queries
    :param fl: The fields to return
    :param sort: The sort order
    :param rows: The number of rows to return
    :param start: Set the start index
    :param cursor_mark: Set the cursor mark
    :param grouping: Set the grouping component
    :param def_type: Set the query type
    :param facet_set: Facet counts
    :param json_facet: Json facets
    :param additional_params: Additional parameters to send with the request, e.g. when using subqueries.
    """

    def __init__(
            self,
            q: Optional[str] = None,
            fq: Optional[List[str]] = None,
            fl: Optional[List[str]] = None,
            sort: Optional[List[str]] = None,
            rows: Optional[int] = None,
            start: Optional[int] = None,
            cursor_mark: Optional[str] = None,
            grouping: Optional["GroupingComponent"] = None,
            def_type: Optional["DefType"] = None,
            facet_set: Optional["FacetSetComponent"] = None,
            json_facet: Optional["JsonFacetComponent"] = None,
            stats: Optional["StatsComponent"] = None,
            additional_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        pass

    async def execute(
            self, context: "SolrServerContext", collection: str
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to query
        """

    async def execute_raw(
            self, context: "SolrServerContext", collection: str
    ) -> Dict[str, Any]:
        """Execute the query, and return the raw solr response as a dictionary.

        :param context: The context for the connection to the solr instance
        :param collection: The collection to query
        """

    def execute_blocking(
            self, context: "SolrServerContext", collection: str
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to query
        """

    def execute_blocking_raw(
            self, context: "SolrServerContext", collection: str
    ) -> Dict[str, Any]:
        """Execute the query, and return the raw solr response as a dictionary.

        :param context: The context for the connection to the solr instance
        :param collection: The collection to query
        """


class CommitType(Enum):
    Hard = "Hard"
    Soft = "Soft"


class UpdateQuery:
    """Builder for an update query

    :param handler: The handler for the update query
    :param commit_type: The commit type for the update query
    """

    def __init__(
            self,
            handler: Optional[str] = "update",
            commit_type: Optional[CommitType] = CommitType.Hard,
    ) -> None:
        pass

    async def execute(
            self, context: "SolrServerContext", collection: str, data: List[Dict[str, Any]]
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to update
        :param data: The data to update
        """

    def execute_blocking(
            self, context: "SolrServerContext", collection: str, data: List[Dict[str, Any]]
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to update
        :param data: The data to update
        """


class DeleteQuery:
    """Builder for a delete query

    :param handler: The handler for the delete query
    :param commit_type: The commit type for the delete query
    """

    def __init__(
            self,
            handler: Optional[str] = "update",
            commit_type: Optional[CommitType] = CommitType.Hard,
            ids: Optional[List[str]] = None,
            queries: Optional[List[str]] = None,
    ) -> None:
        pass

    async def execute(
            self, context: "SolrServerContext", collection: str
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to delete from
        """

    def execute_blocking(
            self, context: "SolrServerContext", collection: str
    ) -> "SolrResponse":
        """Execute the query

        :param context: The context for the connection to the solr instance
        :param collection: The collection to delete from
        """


# endregion

# region clients

class AsyncSolrCloudClient:
    """
    A client for interacting with a SolrCloud cluster asynchronously.

    :param context: The context of the Solr server.

    >>> from solrstice import SolrServerContext, SolrBasicAuth, AsyncSolrCloudClient
    >>> context = SolrServerContext("http://localhost:8983", SolrBasicAuth("user", "pass"))
    >>> client = AsyncSolrCloudClient(context)

    """

    def __init__(self, context: "SolrServerContext"):
        pass

    async def upload_config(
            self, config_name: str, config_path: Union[PathLike[str], str]
    ) -> None:
        """Uploads a Solr config to a Solr instance

        :param config_name: Name of the config
        :param config_path: Path to the config

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def upload_config() -> None:
        ...     await client.upload_config("config_name", "/path/to/config")

        """

    async def get_configs(self) -> List[str]:
        """Gets a list of Solr configs on a Solr instance
        :returns: A list of Solr configs

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> from typing import List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def get_configs() -> List[str]:
        ...     return await client.get_configs()

        """

    async def config_exists(self, config_name: str) -> bool:
        """Checks if a Solr config exists on a Solr instance

        :param config_name: Name of the config
        :returns: True if the config exists, False otherwise

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def config_exists() -> bool:
        ...     return await client.config_exists("config_name")

        """

    async def delete_config(self, config_name: str) -> None:
        """Deletes a Solr config from a Solr instance
        :param config_name: Name of the config

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def delete_config() -> None:
        ...     await client.delete_config("config_name")

        """

    pass

    async def create_collection(
            self,
            name: str,
            config: str,
            shards: Optional[int] = 1,
            replication_factor: Optional[int] = 1,
    ) -> None:
        """
        Create a collection on the Solr server.

        :param name: The name of the collection to create.
        :param config: The name of the config to use for the collection.
        :param shards: The number of shards to create.
        :param replication_factor: The replication factor to use.

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def create_collection() -> None:
        ...     await client.create_collection("collection_name", "config_name", 1, 1)

        """

    async def get_collections(self) -> List[str]:
        """
        Get the list of collections on the Solr server.

        :return: The list of collections on the Solr server.

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> from typing import List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def get_collections() -> List[str]:
        ...     return await client.get_collections()

        """

    async def collection_exists(self, name: str) -> bool:
        """
        Check if a collection exists on the Solr server.

        :param name: The name of the collection to check.
        :return: True if the collection exists, False otherwise.

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def collection_exists() -> bool:
        ...     return await client.collection_exists("collection_name")

        """

    async def delete_collection(self, name: str) -> None:
        """
        Delete a config from the Solr server.
        :param name: The name of the collection to delete.

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def delete_collection() -> None:
        ...     await client.delete_collection("collection_name")

        """

    async def create_alias(self, name: str, collections: List[str]) -> None:
        """
        Create an alias for a collection on the Solr server

        :param name: The name of the alias to create
        :param collections: The collections to alias

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def create_alias() -> None:
        ...     await client.create_alias("alias_name", ["collection_name"])

        """

    async def get_aliases(self) -> Dict[str, List[str]]:
        """
        Get all aliases on the Solr server

        :return: A dictionary of aliases to collections

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> from typing import Dict, List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def get_aliases() -> Dict[str, List[str]]:
        ...     return await client.get_aliases()

        """

    async def alias_exists(self, name: str) -> bool:
        """
        Check if an alias exists on the Solr server

        :param name: The name of the alias to check
        :return: True if the alias exists, False otherwise

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def alias_exists() -> bool:
        ...     return await client.alias_exists("alias_name")

        """

    async def delete_alias(self, name: str) -> None:
        """
        Delete an alias from the Solr server

        :param name: The name of the alias to delete

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def delete_alias() -> None:
        ...     await client.delete_alias("alias_name")

        """

    async def select(self, builder: "SelectQuery", collection: str) -> "SolrResponse":
        """Execute a select query

        Also see :class:`SelectQuery`

        :param builder: The query builder
        :param collection: The collection to query
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient, SelectQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def select() -> SolrResponse:
        ...     query = SelectQuery(q="*:*", rows=10)
        ...     return await client.select(query, "collection_name")

        """

    async def select_raw(self, builder: "SelectQuery", collection: str) -> Dict[str, Any]:
        """Execute a select query. Return raw response from solr

        Also see :class:`SelectQuery`

        :param builder: The query builder
        :param collection: The collection to query
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient, SelectQuery
        >>> from typing import Dict, Any
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def select_raw() -> Dict[str, Any]:
        ...     query = SelectQuery(q="*:*", rows=10)
        ...     return await client.select_raw(query, "collection_name")

        """

    async def index(
            self, builder: "UpdateQuery", collection: str, data: List[Dict[str, Any]]
    ) -> "SolrResponse":
        """Execute an index query

        Also see :class:`UpdateQuery`

        :param builder: The query builder
        :param collection: The collection to index
        :param data: The data to index
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient, UpdateQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def index() -> SolrResponse:
        ...     query = UpdateQuery()
        ...     data_to_index = [{"id": "1", "name": "test"}]
        ...     return await client.index(query, "collection_name", data_to_index)

        """

    async def delete(self, builder: "DeleteQuery", collection: str) -> "SolrResponse":
        """Execute a delete query

        Also see :class:`DeleteQuery`

        :param builder: The query builder
        :param collection: The collection to delete from
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, AsyncSolrCloudClient, DeleteQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = AsyncSolrCloudClient(context)
        >>> async def delete() -> SolrResponse:
        ...     query = DeleteQuery(ids=["1"])
        ...     return await client.delete(query, "collection_name")

        """


class BlockingSolrCloudClient:
    """
    A client for interacting with a SolrCloud cluster non-asynchronously.

    :param context: The context of the Solr server.

    >>> from solrstice import SolrServerContext, SolrBasicAuth, BlockingSolrCloudClient
    >>> context = SolrServerContext("http://localhost:8983", SolrBasicAuth("user", "pass"))
    >>> client = BlockingSolrCloudClient(context)

    """

    def __init__(self, context: "SolrServerContext"):
        pass

    def upload_config(
            self, config_name: str, config_path: Union[PathLike[str], str]
    ) -> None:
        """Uploads a Solr config to a Solr instance

        :param config_name: Name of the config
        :param config_path: Path to the config

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def upload_config() -> None:
        ...     client.upload_config("config_name", "/path/to/config")

        """

    def get_configs(self) -> List[str]:
        """
        Gets a list of Solr configs on a Solr instance
        :returns: A list of Solr configs

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> from typing import List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def get_configs() -> List[str]:
        ...     return client.get_configs()

        """

    def config_exists(self, config_name: str) -> bool:
        """Checks if a Solr config exists on a Solr instance

        :param config_name: Name of the config
        :returns: True if the config exists, False otherwise

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def config_exists() -> bool:
        ...     return client.config_exists("config_name")

        """

    def delete_config(self, config_name: str) -> None:
        """Deletes a Solr config from a Solr instance

        :param config_name: Name of the config

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def delete_config() -> None:
        ...     client.delete_config("config_name")

        """

    def create_collection(
            self,
            name: str,
            config: str,
            shards: Optional[int] = 1,
            replication_factor: Optional[int] = 1,
    ) -> None:
        """
        Create a collection on the Solr server.

        :param name: The name of the collection to create.
        :param config: The name of the config to use for the collection.
        :param shards: The number of shards to create.
        :param replication_factor: The replication factor to use.

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def create_collection() -> None:
        ...    client.create_collection("collection_name", "config_name", 1, 1)

        """

    def get_collections(self) -> List[str]:
        """
        Get the list of collections on the Solr server.

        :return: The list of collections on the Solr server.

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> from typing import List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def get_collections() -> List[str]:
        ...     return client.get_collections()

        """

    def collection_exists(self, name: str) -> bool:
        """
        Check if a collection exists on the Solr server.

        :param name: The name of the collection to check.
        :return: True if the collection exists, False otherwise.

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def collection_exists() -> bool:
        ...     return client.collection_exists("collection_name")

        """

    def delete_collection(self, name: str) -> None:
        """
        Delete a config from the Solr server.

        :param context: The Solr server context.
        :param name: The name of the collection to delete.

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)

        >>> def delete_collection() -> None:
        ...     client.delete_collection("collection_name")

        """

    def create_alias(self, name: str, collections: List[str]) -> None:
        """
        Create an alias for a collection on the Solr server

        :param name: The name of the alias to create
        :param collections: The collections to alias

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def create_alias() -> None:
        ...     client.create_alias("alias_name", ["collection_name"])

        """

    def get_aliases(self) -> Dict[str, List[str]]:
        """
        Get all aliases on the Solr server

        :return: A dictionary of aliases to collections

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> from typing import Dict, List
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def get_aliases() -> Dict[str, List[str]]:
        ...     return client.get_aliases()

        """

    def alias_exists(self, name: str) -> bool:
        """
        Check if an alias exists on the Solr server

        :param name: The name of the alias to check
        :return: True if the alias exists, False otherwise

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def alias_exists() -> bool:
        ...     return client.alias_exists("alias_name")

        """

    def delete_alias(self, name: str) -> None:
        """
        Delete an alias from the Solr server

        :param name: The name of the alias to delete

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def delete_alias() -> None:
        ...     client.delete_alias("alias_name")

        """

    def select(self, builder: "SelectQuery", collection: str) -> "SolrResponse":
        """Execute a select query

        Also see :class:`SelectQuery`

        :param builder: The query builder
        :param collection: The collection to query
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient, SelectQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def select() -> SolrResponse:
        ...     query = SelectQuery(q="*:*", rows=10)
        ...     return client.select(query, "collection_name")

        """

    def select_raw(self, builder: "SelectQuery", collection: str) -> Dict[str, Any]:
        """Execute a select query. Return the raw response from Solr

        Also see :class:`SelectQuery`

        :param builder: The query builder
        :param collection: The collection to query
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient, SelectQuery
        >>> from typing import Dict, Any
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def select_raw() -> Dict[str, Any]:
        ...     query = SelectQuery(q="*:*", rows=10)
        ...     return client.select_raw(query, "collection_name")

        """

    def index(
            self, builder: "UpdateQuery", collection: str, data: List[Dict[str, Any]]
    ) -> "SolrResponse":
        """Execute an index query

        Also see :class:`UpdateQuery`

        :param builder: The query builder
        :param collection: The collection to index
        :param data: The data to index

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient, UpdateQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def index() -> SolrResponse:
        ...     query = UpdateQuery()
        ...     data_to_index = [{"id": "1", "name": "test"}]
        ...     return  client.index(query, "collection_name", data_to_index)

        """

    def delete(self, builder: "DeleteQuery", collection: str) -> "SolrResponse":
        """Execute a delete query

        :param builder: The query builder
        :param collection: The collection to delete from
        :returns: The response from the query

        >>> from solrstice import SolrServerContext, BlockingSolrCloudClient, DeleteQuery
        >>> from solrstice.models import SolrResponse
        >>> context = SolrServerContext("http://localhost:8983")
        >>> client = BlockingSolrCloudClient(context)
        >>> def delete() -> SolrResponse:
        ...     query = DeleteQuery(ids=["1"])
        ...     return client.delete(query, "collection_name")

        """

# endregion



import json
import os
import time
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from urllib.error import HTTPError
from urllib.request import urlopen

from dotenv import load_dotenv

from solrstice import (
    AsyncSolrCloudClient,
    OffLoggingPolicy,
    SolrBasicAuth,
    SolrServerContext,
    SolrSingleServerHost,
    UpdateQuery,
)
from solrstice.collection import create_collection, delete_collection
from solrstice.config import delete_config, upload_config


@dataclass
class Config:
    solr_host: str
    speedbump_host: Optional[str]
    error_nginx_host: Optional[str]
    solr_username: Optional[str]
    solr_password: Optional[str]
    solr_auth: Optional[SolrBasicAuth]
    context: SolrServerContext
    config_path: str
    async_client: AsyncSolrCloudClient


def get_path_prefix() -> str:
    iterations = 0
    path_prefix = ""
    while True:
        if not os.path.exists(os.path.join(path_prefix, "test_setup/.env")):
            path_prefix += "../"
        else:
            break
        iterations += 1
        if iterations > 100:
            raise FileNotFoundError("Could not find test_setup/.env")
    return path_prefix


def create_config(logging: bool = False) -> Config:
    path = os.path.join(get_path_prefix(), "test_setup/.env")
    load_dotenv(path)
    solr_auth = None
    solr_username = os.getenv("SOLR_USERNAME")
    solr_password = os.getenv("SOLR_PASSWORD")
    if solr_username:
        solr_auth = SolrBasicAuth(solr_username, solr_password)
    host = os.getenv("SOLR_HOST")
    assert host is not None
    speedbump_host = os.getenv("SPEEDBUMP_HOST")
    error_nginx_host = os.getenv("ERROR_NGINX_HOST")
    solr_host = SolrSingleServerHost(host)
    context = SolrServerContext(
        solr_host, solr_auth, OffLoggingPolicy() if not logging else None
    )
    wait_for_solr(host, 30)
    return Config(
        host,
        speedbump_host,
        error_nginx_host,
        solr_username,
        solr_password,
        solr_auth,
        context,
        os.path.join(get_path_prefix(), "test_setup/test_collection"),
        AsyncSolrCloudClient(context),
    )


def wait_for_solr(host: str, max_time: int) -> None:
    end = time.time() + max_time
    while time.time() < end:
        try:
            with urlopen(
                f'{host}{"/solr/admin/collections"}?action=CLUSTERSTATUS'
            ) as response:
                if response.status == 200:
                    return
        except HTTPError as e:
            if e.code == 401:
                return
        except Exception:
            pass
        time.sleep(1)
    raise RuntimeError(f"Solr did not respond within {max_time} seconds")


@dataclass
class ErrorTestsSetup:
    error_nginx_host: str
    context: SolrServerContext
    async_client: AsyncSolrCloudClient


def create_nginx_error_config() -> ErrorTestsSetup:
    path = os.path.join(get_path_prefix(), "test_setup/.env")
    load_dotenv(path)
    error_nginx_host = os.getenv("ERROR_NGINX_HOST")
    assert error_nginx_host is not None
    context = SolrServerContext(SolrSingleServerHost(error_nginx_host))
    wait_for_error_nginx(error_nginx_host, 30)
    return ErrorTestsSetup(error_nginx_host, context, AsyncSolrCloudClient(context))


def wait_for_error_nginx(host: str, max_time: int) -> None:
    end = time.time() + max_time
    while time.time() < end:
        try:
            with urlopen(f'{host}{"/status"}') as response:
                if response.status == 200:
                    return
        except Exception:
            pass
        time.sleep(1)
    raise RuntimeError(f"Error nginx did not respond within {max_time} seconds")


@dataclass
class Population:
    id: str
    age: int
    count: int
    interests: List[str]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "Population":
        return Population(
            data["id"],
            data["age"],
            data["count"],
            data["interests"] if "interests" in data else [],
        )

    @staticmethod
    def to_dict(population: "Population") -> Dict[str, Any]:
        return {
            "id": population.id,
            "age": population.age,
            "count": population.count,
            "interests": population.interests,
        }


@dataclass
class City:
    id: str
    city_name: str
    population: List[Population]

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "City":
        return City(
            data["id"],
            data["city_name"],
            (
                [Population.from_dict(x) for x in data["population"]]
                if "population" in data
                else []
            ),
        )

    @staticmethod
    def to_dict(city: "City") -> Dict[str, Any]:
        return {
            "id": city.id,
            "city_name": city.city_name,
            "population": [Population.to_dict(x) for x in city.population],
        }


def load_test_data() -> List[City]:
    with open(os.path.join(get_path_prefix(), "test_setup/test_data.json")) as f:
        return [City.from_dict(x) for x in json.loads(f.read())]


async def index_test_data(context: SolrServerContext, name: str) -> None:
    data = load_test_data()
    update_builder = UpdateQuery()
    await update_builder.execute(context, name, [City.to_dict(x) for x in data])


async def setup_collection(
    context: SolrServerContext, name: str, config_path: str
) -> None:
    try:
        await delete_collection(context, name)
    except RuntimeError:
        pass
    try:
        await delete_config(context, name)
    except RuntimeError:
        pass
    await upload_config(
        context,
        name,
        config_path,
    )
    await create_collection(context, name, name, 1, 1)


async def teardown_collection(context: SolrServerContext, name: str) -> None:
    try:
        await delete_collection(context, name)
    except RuntimeError:
        pass
    try:
        await delete_config(context, name)
    except RuntimeError:
        pass



from ._solrstice import (
    AsyncSolrCloudClient,
    BlockingSolrCloudClient,
    CommitType,
    DefType,
    DeleteQuery,
    DismaxQuery,
    EdismaxQuery,
    FacetSetComponent,
    StatsComponent,
    FastLoggingPolicy,
    FieldFacetComponent,
    FieldFacetEntry,
    FieldFacetMethod,
    FieldFacetSort,
    GroupFormatting,
    GroupingComponent,
    JsonFacetComponent,
    JsonFacetType,
    JsonFacetSortDirection,
    JsonQueryFacet,
    JsonStatFacet,
    JsonTermsFacet,
    JsonTermsFacetMethod,
    LoggingPolicy,
    LuceneQuery,
    OffLoggingPolicy,
    PivotFacetComponent,
    PrettyLoggingPolicy,
    QueryOperator,
    SelectQuery,
    SolrAuth,
    SolrBasicAuth,
    SolrHost,
    SolrMultipleServerHost,
    SolrServerContext,
    SolrSingleServerHost,
    UpdateQuery,
    ZookeeperEnsembleHost,
    ZookeeperEnsembleHostConnector,
)

__all__ = [
    "SolrAuth",
    "SolrBasicAuth",
    "QueryOperator",
    "DefType",
    "LuceneQuery",
    "DismaxQuery",
    "EdismaxQuery",
    "FacetSetComponent",
    "StatsComponent",
    "PivotFacetComponent",
    "FieldFacetComponent",
    "FieldFacetSort",
    "FieldFacetMethod",
    "FieldFacetEntry",
    "JsonFacetComponent",
    "JsonFacetType",
    "JsonFacetSortDirection",
    "JsonTermsFacet",
    "JsonTermsFacetMethod",
    "JsonQueryFacet",
    "JsonStatFacet",
    "SolrHost",
    "SolrSingleServerHost",
    "SolrMultipleServerHost",
    "ZookeeperEnsembleHost",
    "ZookeeperEnsembleHostConnector",
    "LoggingPolicy",
    "OffLoggingPolicy",
    "FastLoggingPolicy",
    "PrettyLoggingPolicy",
    "SolrServerContext",
    "GroupFormatting",
    "GroupingComponent",
    "SelectQuery",
    "CommitType",
    "UpdateQuery",
    "DeleteQuery",
    "AsyncSolrCloudClient",
    "BlockingSolrCloudClient",
]



from typing import Generator

import pytest

from solrstice import (
    FacetSetComponent,
    FieldFacetComponent,
    FieldFacetEntry,
    PivotFacetComponent,
    SelectQuery,
)

from .helpers import (
    Config,
    create_config,
    index_test_data,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_facet_pivot_works(config: Config) -> None:
    name = "FacetPivot"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        pivot = PivotFacetComponent(["interests,age"])
        facet_set = FacetSetComponent(pivots=pivot)
        select_builder = SelectQuery(facet_set=facet_set)
        response = await config.async_client.select(select_builder, name)
        facets = response.get_facet_set()
        pivot_result = facets.get_pivots()
        interests_age = pivot_result["interests,age"]
        cars_pivot = next(p for p in interests_age if p.get_value() == "cars")
        assert cars_pivot.get_count() == 1
        age_pivot = cars_pivot.get_pivots()[0]
        assert age_pivot.get_value() == 20
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_facet_query_works(config: Config) -> None:
    name = "FacetQuery"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        facet_set = FacetSetComponent(queries=["age:[0 TO 59]"])
        select_builder = SelectQuery(facet_set=facet_set)
        response = await config.async_client.select(select_builder, name)
        facets = response.get_facet_set()
        queries = facets.get_queries()
        query = queries.get("age:[0 TO 59]")
        assert query == 4
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_facet_field_works(config: Config) -> None:
    name = "FacetField"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        fields = FieldFacetComponent(fields=[FieldFacetEntry("age")])
        facet_set = FacetSetComponent(fields=fields)
        select_builder = SelectQuery(facet_set=facet_set)
        response = await config.async_client.select(select_builder, name)
        facets = response.get_facet_set()
        fields_result = facets.get_fields()
        age = fields_result["age"]
        assert len(age) == 3
    finally:
        await teardown_collection(config.context, name)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Optional

import pytest
from dotenv import load_dotenv

from solrstice import (
    SolrBasicAuth,
    SolrMultipleServerHost,
    SolrServerContext,
    SolrSingleServerHost,
    ZookeeperEnsembleHostConnector,
)
from solrstice.config import get_configs, get_configs_blocking

from .helpers import wait_for_solr


@dataclass
class Config:
    host: str
    zookeeper_host: str
    auth: Optional[SolrBasicAuth]


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    path = Path("../../test_setup/.env").resolve()
    load_dotenv(path)
    solr_auth = None
    solr_username = os.getenv("SOLR_USERNAME")
    solr_password = os.getenv("SOLR_PASSWORD")

    if solr_username is not None and solr_password != "":
        solr_auth = SolrBasicAuth(
            solr_username,
            solr_password,
        )
    host = os.getenv("SOLR_HOST")
    assert host is not None
    zookeeper_host = os.getenv("ZK_HOST")
    assert zookeeper_host is not None

    yield Config(host, zookeeper_host, solr_auth)


@pytest.mark.asyncio
async def test_zookeeper_connection_works(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(
        await ZookeeperEnsembleHostConnector([config.zookeeper_host], 30).connect(),
        config.auth,
    )
    await get_configs(context)


def test_zookeeper_connection_works_blocking(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(
        ZookeeperEnsembleHostConnector([config.zookeeper_host], 30).connect_blocking(),
        config.auth,
    )
    get_configs_blocking(context)


@pytest.mark.asyncio
async def test_solr_single_server_works(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(SolrSingleServerHost(config.host), config.auth)
    await get_configs(context)


@pytest.mark.asyncio
async def test_solr_single_server_works_with_string(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(config.host, config.auth)
    await get_configs(context)


@pytest.mark.asyncio
async def test_multiple_server_works(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(SolrMultipleServerHost([config.host], 5), config.auth)
    await get_configs(context)


def test_solr_multiple_server_works_blocking(config: Config) -> None:
    wait_for_solr(config.host, 30)
    context = SolrServerContext(SolrMultipleServerHost([config.host], 5), config.auth)
    get_configs_blocking(context)



# Test that the GIL is not blocked while waiting for a response.
import asyncio
import time
from multiprocessing.pool import ThreadPool
from typing import Generator, List, Optional

import pytest

from solrstice import (
    AsyncSolrCloudClient,
    BlockingSolrCloudClient,
    SolrBasicAuth,
    SolrServerContext,
    SolrSingleServerHost,
)

from .helpers import Config, create_config


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


def test_blocking_client_does_not_block_gil_config(config: Config) -> None:
    processes = 8
    if not config.speedbump_host:
        pytest.skip("No speedbump host configured")
    with ThreadPool(processes) as pool:
        start_time = time.time()
        tasks = [
            pool.apply_async(
                get_configs_blocking,
                (config.speedbump_host, config.solr_username, config.solr_password),
            )
            for _ in range(processes)
        ]
        for task in tasks:
            task.get(timeout=processes * 3)

    elapsed_seconds = time.time() - start_time
    # Presume since the latency is 2s, that it should have completed in max 1s, given no GIL blocking.
    assert elapsed_seconds < processes


@pytest.mark.asyncio
async def test_async_client_does_not_block_event_loop(config: Config) -> None:
    processes = 8
    if not config.speedbump_host:
        pytest.skip("No speedbump host configured")
    start_time = time.time()
    tasks = [
        asyncio.create_task(
            get_configs_async(
                config.speedbump_host, config.solr_username, config.solr_password
            )
        )
        for _ in range(processes)
    ]
    await asyncio.gather(*tasks)
    elapsed_seconds = time.time() - start_time
    # Presume since the latency is 2s, that it should have completed in max 1s, given no GIL blocking.
    assert elapsed_seconds < processes


def get_configs_blocking(
    host: str, username: Optional[str], password: Optional[str]
) -> List[str]:
    auth = None if not username else SolrBasicAuth(username, password)
    client = BlockingSolrCloudClient(
        SolrServerContext(SolrSingleServerHost(host), auth)
    )
    return client.get_configs()


async def get_configs_async(
    host: str, username: Optional[str], password: Optional[str]
) -> List[str]:
    auth = None if not username else SolrBasicAuth(username, password)
    client = AsyncSolrCloudClient(SolrServerContext(SolrSingleServerHost(host), auth))
    return await client.get_configs()



from typing import Generator

import pytest

from solrstice import (
    JsonFacetComponent,
    JsonQueryFacet,
    JsonStatFacet,
    JsonTermsFacet,
    SelectQuery, JsonTermsFacetMethod,
)

from .helpers import (
    Config,
    create_config,
    index_test_data,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_json_query_facet_works(config: Config) -> None:
    name = "JsonQueryFacet"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(
            json_facet=JsonFacetComponent(
                facets={"below_60": JsonQueryFacet(q="age:[0 TO 59]")}
            )
        )
        response = await config.async_client.select(select_builder, name)
        facets = response.get_json_facets()
        assert facets is not None
        below_60 = facets.get_nested_facets()["below_60"]
        assert below_60.get_count() == 4
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_json_terms_facet_works(config: Config) -> None:
    name = "JsonTermsFacet"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(
            json_facet=JsonFacetComponent(facets={"age": JsonTermsFacet("age")})
        )
        response = await config.async_client.select(select_builder, name)
        facets = response.get_json_facets()
        assert facets is not None
        age_buckets = facets.get_nested_facets()["age"].get_buckets()
        assert len(age_buckets) == 3
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_json_terms_facet_extended(config: Config) -> None:
    name = "JsonTermsFacetExtended"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(
            json_facet=JsonFacetComponent(
                facets={
                    "city_name": JsonTermsFacet(
                        "city_name",
                        num_buckets=True,
                        all_buckets=True,
                        missing=True,
                        limit=10,
                        prefix="A",
                        method=JsonTermsFacetMethod.DocValues
                    )
                })
        )
        response = await config.async_client.select(select_builder, name)
        facets = response.get_json_facets()
        assert facets is not None
        city_name = facets.get_nested_facets()["city_name"]
        assert len(city_name.get_buckets()) == 1
        assert city_name.get_missing() is not None
        assert city_name.get_all_buckets() is not None
        assert city_name.get_num_buckets() == 1
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_json_stat_facet_works(config: Config) -> None:
    name = "JsonStatFacet"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(
            json_facet=JsonFacetComponent(
                facets={"total_people": JsonStatFacet("sum(count)")}
            )
        )
        response = await config.async_client.select(select_builder, name)
        facets = response.get_json_facets()
        assert facets is not None
        total_people = facets.get_flat_facets()["total_people"]
        assert total_people == 1000
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_json_facet_sub_works(config: Config) -> None:
    name = "JsonFacetSub"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        select_builder = SelectQuery(
            json_facet=JsonFacetComponent(
                facets={
                    "below_60": JsonQueryFacet(
                        q="age:[0 TO 59]",
                        facets={"total_people": JsonStatFacet("sum(count)")},
                    )
                }
            )
        )
        response = await config.async_client.select(select_builder, name)
        facets = response.get_json_facets()
        assert facets is not None
        total_people = facets.get_nested_facets()["below_60"].get_flat_facets()[
            "total_people"
        ]
        assert total_people == 750.0
    finally:
        await teardown_collection(config.context, name)



from typing import Generator

import pytest

from solrstice import SelectQuery

from .helpers import ErrorTestsSetup, create_nginx_error_config


@pytest.fixture()
def config() -> Generator[ErrorTestsSetup, None, None]:
    yield create_nginx_error_config()


@pytest.mark.asyncio
async def test_sensible_error_message_if_not_solr_server(
    config: ErrorTestsSetup,
) -> None:
    try:
        await config.async_client.select(SelectQuery(), "error_collection")
    except Exception as e:
        assert "500" in str(e)


@pytest.mark.asyncio
async def test_sensible_error_message_if_non_existent_collection(
    config: ErrorTestsSetup,
) -> None:
    try:
        await config.async_client.select(SelectQuery(), "notfound_collection")
    except Exception as e:
        assert "404" in str(e)


@pytest.mark.asyncio
async def test_sensible_error_message_if_200_but_not_solr(
    config: ErrorTestsSetup,
) -> None:
    try:
        await config.async_client.select(SelectQuery(), "always_200")
    except Exception as e:
        assert "200" in str(e)



from os import PathLike
from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    from solrstice import SolrServerContext

    Somepath = Union[PathLike[str], str]
else:
    Somepath = Union[PathLike, str]

async def upload_config(
    context: "SolrServerContext", config_name: str, config_path: Somepath
) -> None:
    """Uploads a Solr config to a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    :param config_path: Path to the config
    """
    pass

def upload_config_blocking(
    context: "SolrServerContext", config_name: str, config_path: Somepath
) -> None:
    """Uploads a Solr config to a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    :param config_path: Path to the config
    """
    pass

async def delete_config(context: "SolrServerContext", config_name: str) -> None:
    """Deletes a Solr config from a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    """
    pass

def delete_config_blocking(context: "SolrServerContext", config_name: str) -> None:
    """Deletes a Solr config from a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    """
    pass

async def config_exists(context: "SolrServerContext", config_name: str) -> bool:
    """Checks if a Solr config exists on a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    """
    pass

def config_exists_blocking(context: "SolrServerContext", config_name: str) -> bool:
    """Checks if a Solr config exists on a Solr instance

    :param context: SolrServerRequest context
    :param config_name: Name of the config
    """
    pass

async def get_configs(context: "SolrServerContext") -> List[str]:
    """Gets a list of Solr configs on a Solr instance

    :param context: SolrServerRequest context
    """
    pass

def get_configs_blocking(context: "SolrServerContext") -> List[str]:
    """Gets a list of Solr configs on a Solr instance

    :param context: SolrServerRequest builder
    """
    pass

__all__ = [
    "upload_config",
    "upload_config_blocking",
    "delete_config",
    "delete_config_blocking",
    "config_exists",
    "config_exists_blocking",
    "get_configs",
    "get_configs_blocking",
]



from typing import Generator

import pytest

from solrstice import GroupingComponent, SelectQuery

from .helpers import (
    Config,
    create_config,
    index_test_data,
    setup_collection,
    teardown_collection,
    wait_for_solr,
)


@pytest.fixture()
def config() -> Generator[Config, None, None]:
    yield create_config()


@pytest.mark.asyncio
async def test_get_group_field_result_works(config: Config) -> None:
    name = "GroupFieldQuery"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        group_builder = GroupingComponent(fields=["age"], limit=10)
        select_builder = SelectQuery(fq=["age:[* TO *]"], grouping=group_builder)
        groups = (await select_builder.execute(config.context, name)).get_groups()
        age_group = groups["age"]
        group = age_group.get_field_result()
        assert group is not None
        assert age_group.get_n_groups() is None
        assert age_group.get_matches() > 0
        assert len(group) > 0
    finally:
        await teardown_collection(config.context, name)


@pytest.mark.asyncio
async def test_get_group_query_result_works(config: Config) -> None:
    name = "GroupQueryQuery"
    wait_for_solr(config.solr_host, 30)

    try:
        await setup_collection(config.context, name, config.config_path)
        await index_test_data(config.context, name)

        group_builder = GroupingComponent(
            queries=["age:[0 TO 59]", "age:[60 TO *]"], limit=10
        )
        select_builder = SelectQuery(fq=["age:[* TO *]"], grouping=group_builder)
        groups = (await select_builder.execute(config.context, name)).get_groups()
        age_group = groups["age:[0 TO 59]"]
        group = age_group.get_query_result()
        assert group is not None
        assert len(group.get_docs()) > 0
    finally:
        await teardown_collection(config.context, name)



from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from solrstice import SolrServerContext

async def create_alias(
    context: "SolrServerContext", name: str, collections: List[str]
) -> None:
    """
    Create an alias for a collection on the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to create
    :param collections: The collections to alias
    """

def create_alias_blocking(
    context: "SolrServerContext", name: str, collections: List[str]
) -> None:
    """
    Create an alias for a collection on the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to create
    :param collections: The collections to alias
    """

async def get_aliases(context: "SolrServerContext") -> Dict[str, List[str]]:
    """
    Get all aliases on the Solr server

    :param context: The Solr server context
    :return: A dictionary of aliases to collections
    """

def get_aliases_blocking(context: "SolrServerContext") -> Dict[str, List[str]]:
    """
    Get all aliases on the Solr server

    :param context: The Solr server context
    :return: A dictionary of aliases to collections
    """

async def alias_exists(context: "SolrServerContext", name: str) -> bool:
    """
    Check if an alias exists on the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to check
    :return: True if the alias exists, False otherwise
    """

def alias_exists_blocking(context: "SolrServerContext", name: str) -> bool:
    """
    Check if an alias exists on the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to check
    :return: True if the alias exists, False otherwise
    """

async def delete_alias(context: "SolrServerContext", name: str) -> None:
    """
    Delete an alias from the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to delete
    """

def delete_alias_blocking(context: "SolrServerContext", name: str) -> None:
    """
    Delete an alias from the Solr server

    :param context: The Solr server context
    :param name: The name of the alias to delete
    """

__all__ = [
    "create_alias",
    "create_alias_blocking",
    "get_aliases",
    "get_aliases_blocking",
    "alias_exists",
    "alias_exists_blocking",
    "delete_alias",
    "delete_alias_blocking",
]
